#!/usr/bin/env python3
"""
Javantea's Reverse Engineering Tool GTK interface
by Javantea
Dec 9, 2014

Based on http://python-gtk-3-tutorial.readthedocs.org/en/latest/introduction.html
Very similar to my work on mail1gtk which used PyGTK.

Memory usage on a 1.2MB file:
jvoss    29188 31.8  0.6 1350920 97812 pts/30  Sl   11:59   1:35 /usr/bin/python2.7 gtk-javre.py
Quite alright.
Memory usage on a 112kB file:
jvoss    21043 21.2  0.4 1327700 64592 pts/30  Sl   12:06   0:05 /usr/bin/python2.7 gtk-javre.py
Quite fine.

CPU time disassembling /bin/ls: 5 seconds.
CPU time disassembling /usr/bin/qcomicbook: 1 minute 35 seconds.
Definitely could be much much better (see objdump) but it's a start.

TODO: Add verbosity.
"""
from __future__ import print_function
from gi.repository import Gtk, Pango
import disasm1

def doNothing(sect_name, sect_disasm, dynsym_addrs, dynsym_names, elf_st, sections, progs, data, get_pc_thunk_bx):
	"""
	disasm1.disassemble wants a function like this. Here it is.
	"""
	pass
#end def doNothing(sect_name, sect_disasm, dynsym_addrs, dynsym_names, elf_st, sections, progs, data, get_pc_thunk_bx)


def disasmToString(sect_name, sect_disasm, dynsym_addrs, dynsym_names, elf_st, sections, progs, data, get_pc_thunk_bx):
	"""
	Based on printDisassembly
	"""
	ret = sect_name + ':\n'
	functions = []
	for instr in sect_disasm:
		if instr[2][:7] == 'CALL 0x':
			addr = int(instr[2][7:], 16)
			functions.append(addr)
		#end if
	#next instr
	
	for instr in sect_disasm:
		if instr[0] in dynsym_addrs:
			func_name = dynsym_names[dynsym_addrs.index(instr[0])]
			ret += (func_name + ':\n')
		elif instr[0] in functions:
			ret += ('sub_%x:' % instr[0]) + '\n'
		#end if
		ret += (('%8x' % instr[0]).rjust(8, ' ') + ': ')
		if instr[2][:7] == 'CALL 0x':
			func_name = None
			addr = int(instr[2][7:], 16)
			if addr in dynsym_addrs:
				func_name = dynsym_names[dynsym_addrs.index(addr)]
			#end if
			if func_name:
				ret += ("CALL " + instr[2][7:] + '<' + func_name + '>\n')
			else:
				ret += ("CALL " + instr[2][7:] + '<sub_' + instr[2][7:] + '>\n')
			#end if
		else:
			ret += (instr[2]) + '\n'
		#end if
	#next instr
	return ret
#end def disasmToString(sect_name, sect_disasm, dynsym_addrs, dynsym_names, elf_st, sections, progs, data, get_pc_thunk_bx)

def disasmToBuffer(textBuffer, sect_name, sect_disasm, dynsym_addrs, dynsym_names, elf_st, sections, progs, data, get_pc_thunk_bx):
	"""
	Based on disasmToString
	"""
	start_iter = textBuffer.get_end_iter()
	curr_iter = start_iter
	textBuffer.insert_with_tags_by_name(curr_iter, sect_name + ':\n', 'bold')
	functions = []
	for instr in sect_disasm:
		if instr[2][:7] == 'CALL 0x':
			addr = int(instr[2][7:], 16)
			functions.append(addr)
		#end if
	#next instr
	endCurlyBraces = []
	prevTest = None
	prevCmp  = None
	prevEBX  = None
	currentFunction = 0
	
	for i, instr in enumerate(sect_disasm):
		if instr[0] in dynsym_addrs:
			func_name = dynsym_names[dynsym_addrs.index(instr[0])]
			textBuffer.insert_with_tags_by_name(curr_iter, func_name + ':\n', 'bold')
			currentFunction = i
		elif instr[0] in functions:
			textBuffer.insert_with_tags_by_name(curr_iter, 'sub_%x:\n' % instr[0], 'bold')
			currentFunction = i
		#end if
		if instr[0] in endCurlyBraces:
			textBuffer.insert_with_tags_by_name(curr_iter, '; }\n', 'c_code')
		#end if
		textBuffer.insert_with_tags_by_name(curr_iter, ('%8x' % instr[0]).rjust(8, ' ') + ': ', 'address')
		if instr[2][:7] == 'CALL 0x':
			func_name = None
			addr = int(instr[2][7:], 16)
			if addr in dynsym_addrs:
				func_name = dynsym_names[dynsym_addrs.index(addr)]
			#end if
			is_pc_thunk_bx = False
			add = ''
			if addr == get_pc_thunk_bx:
				# EBX is now at the next instruction.
				prevEBX = int(instr[0] + instr[1])
				flattenEBX = False
				add = ' ; EBX = ' + hex(prevEBX)
				is_pc_thunk_bx = True
			#end if
			end = '\n'
			if add: end = ''
			if func_name:
				textBuffer.insert_with_tags_by_name(curr_iter, "CALL ", 'call_op')
				textBuffer.insert_with_tags_by_name(curr_iter, instr[2][7:] + ' <', 'instruction')
				textBuffer.insert_with_tags_by_name(curr_iter, func_name, 'bold')
				textBuffer.insert_with_tags_by_name(curr_iter, '>' + end, 'instruction')
				if add:
					textBuffer.insert_with_tags_by_name(curr_iter, add + '\n', 'c_code')
				#end if
			elif is_pc_thunk_bx:
				textBuffer.insert_with_tags_by_name(curr_iter, "CALL ", 'call_op')
				textBuffer.insert_with_tags_by_name(curr_iter, instr[2][7:] + ' <', 'instruction')
				textBuffer.insert_with_tags_by_name(curr_iter, __x86.get_pc_thunk.bx, 'bold')
				textBuffer.insert_with_tags_by_name(curr_iter, '>' + end, 'instruction')
				if add:
					textBuffer.insert_with_tags_by_name(curr_iter, add + '\n', 'c_code')
				#end if
			else:
				textBuffer.insert_with_tags_by_name(curr_iter, "CALL ", 'call_op')
				textBuffer.insert_with_tags_by_name(curr_iter, instr[2][7:] + ' <', 'instruction')
				textBuffer.insert_with_tags_by_name(curr_iter, 'sub_' + instr[2][7:], 'bold')
				textBuffer.insert_with_tags_by_name(curr_iter, '>' + end, 'instruction')
				if add:
					textBuffer.insert_with_tags_by_name(curr_iter, add + '\n', 'c_code')
				#end if
			#end if
		elif instr[2][:4] == 'MOV ':
			z = ''
			if ', 0x' in instr[2]:
				addr = int(instr[2][instr[2].index(', 0x')+4:], 16)
				z = None
				if addr > 0 and addr < 256:
					z = chr(addr)
				elif addr == 0:
					z = None
				else:
					z = disasm1.deref(addr, data, elf_st, sections, progs)
				#end if
				#if z == None:
				#end if
				if z != None:
					z = ' ; ' + repr(z)
				else:
					z = ''
				#end if
			#end if
			if z == None:
				z = disasm1.InstrToC(instr[2])
				if z == None:
					z = ''
				else:
					python_object = z[1]
					z = ' ; ' + z[0]
				#end if
				# non-verbose throws away data.
				z = ''
			#end if
			textBuffer.insert_with_tags_by_name(curr_iter, instr[2], 'instruction')
			if z != '':
				textBuffer.insert_with_tags_by_name(curr_iter, z, 'string')
			#end if
			textBuffer.insert_with_tags_by_name(curr_iter, '\n', 'instruction')
		elif instr[2][:5] == 'TEST ':
			prevTest = instr
			prevCmp  = None
			textBuffer.insert_with_tags_by_name(curr_iter, instr[2], 'instruction')
			textBuffer.insert_with_tags_by_name(curr_iter, ';\n', 'comment')
		elif instr[2][:4] == 'CMP ':
			prevCmp  = instr
			prevTest = None
			textBuffer.insert_with_tags_by_name(curr_iter, instr[2], 'instruction')
			textBuffer.insert_with_tags_by_name(curr_iter, ';\n', 'comment')
		elif instr[2][:1] == 'J' and instr[2][:3] != 'JMP':
			isIf = True
			isLoop = False
			dsth = instr[2].split(' ', 1)[1]
			dsti = 0
			if dsth[:2] == '0x':
				dsti = int(dsth, 16)
				if dsti <= instr[0]:
					# It's a doggone loop!
					isLoop = True
					isIf = False
				#end if
			#end if
			textBuffer.insert_with_tags_by_name(curr_iter, instr[2], 'instruction')
			if prevCmp or prevTest:
				prevX = prevCmp or prevTest
				if isIf:
					textBuffer.insert_with_tags_by_name(curr_iter, ' ; if(' + disasm1.cmpToC(prevX,  instr[2]) + ') {\n', 'c_code')
				else:
					textBuffer.insert_with_tags_by_name(curr_iter, ' ; } while(' + disasm1.cmpToC(prevX,  instr[2], False) + ')\n', 'c_code')
				#end if
			else:
				textBuffer.insert_with_tags_by_name(curr_iter, '\n', 'instruction')
			#end if
			prevCmp  = None
			prevTest = None
			s = instr[2].split(' ', 1)[1]
			if isIf:
				endCurlyBraces.append(dsti)
			#end if
		elif instr[2][:4] == 'LEA ':
			# LEA RAX, [RIP+0x204d6d]
			if '[RIP+' not in instr[2]:
				if (not prevEBX) or ('[EBX' not in instr[2]):
					textBuffer.insert_with_tags_by_name(curr_iter, instr[2] + '\n', 'instruction')
					continue
				#end if
			#end if
			textBuffer.insert_with_tags_by_name(curr_iter, instr[2] + ' ', 'instruction')
			if '[RIP+' in instr[2]:
				rpos = instr[2].index('[RIP+')
				addr = int(instr[2][rpos+5:instr[2].index(']', rpos+5)], 16) + instr[0] + instr[1]
			else:
				# NOTE: This assumes we didn't jump here with a different ebx.
				rpos = instr[2].index('[EBX')
				addr = int(instr[2][rpos+4:instr[2].index(']', rpos+4)], 16) + prevEBX
			#end if
			z = disasm1.deref(addr, data, elf_st, sections, progs)
			textBuffer.insert_with_tags_by_name(curr_iter, '; ' + hex(int(addr)) + ' ' + repr(z) + '\n', 'comment')
		elif instr[2][:3] == 'RET':
			# Find the return value of the current function.
			cr = disasm1.findConstantReturn(sect_name, sect_disasm, currentFunction)
			z = ''
			end = '\n'
			if len(cr) == 2 and cr[0] != None:
				z = ' ; return ' + hex(cr[0]) + ";"
				end = ''
			#end if
			textBuffer.insert_with_tags_by_name(curr_iter, instr[2] + end, 'instruction')
			if z:
				textBuffer.insert_with_tags_by_name(curr_iter, z + '\n', 'c_code')
			#end if
		else:
			z = disasm1.InstrToC(instr[2])
			if z == None:
				z = ''
			else:
				python_object = z[1]
				z = ' ; ' + z[0]
				if prevEBX != None:
					newEBX = python_object.stepEBX(prevEBX)
					if newEBX != prevEBX:
						z += ' ; EBX = ' + hex(newEBX)
					#end if
					prevEBX = newEBX
					flattenEBX = False
				#end if
			#end if
			# non-verbose throws away data.
			z = ''
			textBuffer.insert_with_tags_by_name(curr_iter, instr[2] + '\n', 'instruction')
		#end if
	#next instr
	#return ret
#end def disasmToBuffer(sect_name, sect_disasm, dynsym_addrs, dynsym_names, elf_st, sections, progs, data)

class JavREWindow(Gtk.Window):
	"""
	Javantea's Reverse Engineering Tool window
	Contains everything in that window.
	"""
	def __init__(self):
		Gtk.Window.__init__(self, title="JavRE")
		self.set_default_size(800, 600)
		
		self.sections = Gtk.VBox(spacing = 2)
		toolbar = Gtk.HBox(spacing = 2)
		self.textArea = Gtk.TextView()
		self.textArea.set_property('has-tooltip', True)
		self.textArea.connect("query-tooltip", self.textArea_tooltip)
		self.scrolledWindow = Gtk.ScrolledWindow()
		self.scrolledWindow.add(self.textArea)
		self.reverse = Gtk.Button(label="Open")
		self.reverse.connect("clicked", self.on_reverse_clicked)
		self.save = Gtk.Button(label="Save")
		self.save.connect("clicked", self.on_save_clicked)
		self.sections.pack_start(self.scrolledWindow, True, True, 0)
		toolbar.pack_start(self.reverse, False, False, 0)
		toolbar.pack_start(self.save, False, False, 0)
		self.sections.pack_start(toolbar, False, False, 0)
		self.add(self.sections)
	#end def JavREWindow()
	
	def on_reverse_clicked(self, widget):
		fileDialog = Gtk.FileChooserDialog("Open Executable", self, Gtk.FileChooserAction.OPEN, (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
		response = fileDialog.run()
		if response == Gtk.ResponseType.OK:
			filename = fileDialog.get_filename()
			fileDialog.destroy()
			self.on_file_chosen(widget, filename)
		else:
			fileDialog.destroy()
		#end if
	#end def on_reverse_clicked(widget)

	def on_file_chosen(self, widget, filename):
		elf_data = open(filename, 'rb').read()
		disasm_objs = disasm1.disassemble(elf_data, doNothing, print)
		if len(disasm_objs) != 15:
			print("Error: disasm1.disassemble is broken!")
			return
		#end if
		
		(elf_st, sections, progs, section_names, 
			init_section, plt_section, text_section, fini_section, 
			sym_names, sym_addrs,
			init_disasm, plt_disasm, text_disasm, fini_disasm, 
			get_pc_thunk_bx) = disasm_objs
		
		# TODO: more syntax highlighting.
		# TODO: notice when they modify a value.
		x = Gtk.TextBuffer()
		bold_tag = x.create_tag("bold", family="Monospace", weight=Pango.Weight.BOLD)
		normal_tag = x.create_tag("normal", family="Monospace", weight=0)
		addr_tag = x.create_tag("address", family="Monospace", weight=0)
		instr_tag = x.create_tag("instruction", family="Monospace", weight=0)
		instr_tag = x.create_tag("call_op", family="Monospace", foreground="#cc00cc", weight=0)
		string_tag = x.create_tag("string", family="Monospace", foreground="#cc0000", weight=0)
		comment_tag = x.create_tag("comment", family="Monospace", foreground="#00cc00", weight=0)
		c_code_tag = x.create_tag("c_code", family="Monospace", foreground="#0000cc", weight=0)
		if init_disasm:
			disasmToBuffer(x, ".init", init_disasm, sym_addrs, sym_names, elf_st, sections, progs, elf_data, get_pc_thunk_bx)
		#end if
		if plt_disasm:
			disasmToBuffer(x, ".plt",  plt_disasm,  sym_addrs, sym_names, elf_st, sections, progs, elf_data, get_pc_thunk_bx)
		#end if
		if text_disasm:
			disasmToBuffer(x, ".text", text_disasm, sym_addrs, sym_names, elf_st, sections, progs, elf_data, get_pc_thunk_bx)
		#end if
		if fini_disasm:
			disasmToBuffer(x, ".fini", fini_disasm, sym_addrs, sym_names, elf_st, sections, progs, elf_data, get_pc_thunk_bx)
		#end if
		
		#x.set_text(s)
		self.textArea.set_buffer(x)
	#end def on_file_chosen(widget, filename)
	
	def on_save_clicked(self, widget):
		fileDialog = Gtk.FileChooserDialog("Save Disassembly", self, Gtk.FileChooserAction.SAVE, (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_SAVE, Gtk.ResponseType.OK))
		response = fileDialog.run()
		if response == Gtk.ResponseType.OK:
			filename = fileDialog.get_filename()
			fileDialog.destroy()
			self.on_outfile_chosen(widget, filename)
		else:
			fileDialog.destroy()
		#end if
	#end def on_reverse_clicked(widget)
	
	def on_outfile_chosen(self, widget, filename):
		x = self.textArea.get_buffer()
		data = x.get_text(x.get_start_iter(), x.get_end_iter(), True)
		open(filename, 'w').write(data)
	#end def on_outfile_chosen(widget, filename)
	
	def textArea_tooltip(self, widget, x, y, keyboard_mode, tooltip):
		"""
		Give data to the user based on what they are hovering over.
		"""
		# TODO: write this tooltip code. See link below.
		# http://www.bravegnu.org/gtktext/x498.html
		return False
		print('tooltip', x, y, keyboard_mode, tooltip)
		# A float that gives you the position vertically.
		#self.scrolledWindow.get_vadjustment().get_value()
		#visibleRect = widget.get_visible_rect()
		# These seem to be the coordinates we want.
		print('buffer_coords', widget.window_to_buffer_coords(Gtk.TextWindowType.TEXT, x, y))
		return False
	#end def textArea_tooltip(widget)
	
#end class MyWindow(Gtk.Window)

def main():
	import sys
	filename = None
	disasm1.verbose = False
	if len(sys.argv) > 1:
		filename = sys.argv[1]
	#end if
	win = JavREWindow()
	win.connect("delete-event", Gtk.main_quit)
	if filename:
		win.on_file_chosen(None, filename)
	#end if
	win.show_all()
	Gtk.main()
#end def main()

if __name__ == '__main__':
	main()
#end if
